from floraconcierge.mapping.mixins import Collection as CollectionMixin


class Collection(CollectionMixin):

    """
    :type items: list of Object
    """

    def __init__(self, *args, **kwargs):
        self.count = None
        self.items = None
        self.offset = None
        self.total = None

        super(Collection, self).__init__(*args, **kwargs)



from floraconcierge.client.types import Object


class Auth(Object):

    """
    :type id: int
    """

    def __init__(self, *args, **kwargs):
        self.app_id = None
        self.auth_key = None
        self.checkip = None
        self.expire = None
        self.id = None
        self.ip = None
        self.lasttime = None
        self.remember = None
        self.user_id = None

        super(Auth, self).__init__(*args, **kwargs)


class Address(Object):

    """
    :type id: int
    :type name: str
    """

    def __init__(self, *args, **kwargs):
        self.address = None
        self.city = None
        self.city_id = None
        self.country = None
        self.country_id = None
        self.email = None
        self.id = None
        self.name = None
        self.phone = None

        super(Address, self).__init__(*args, **kwargs)



from floraconcierge.client.types import Object


class Language(Object):

    """
    :type id: int
    :type iso: str
    :type name: str
    """

    def __init__(self, *args, **kwargs):
        self.icon = None
        self.id = None
        self.iso = None
        self.locale = None
        self.name = None

        super(Language, self).__init__(*args, **kwargs)


class Snippet(Object):

    """
    :type id: int
    """

    def __init__(self, *args, **kwargs):
        self.application_id = None
        self.content = None
        self.id = None
        self.uid = None

        super(Snippet, self).__init__(*args, **kwargs)


class User(Object):

    """
    :type auth: Auth
    :type id: int
    :type name: str
    """

    def __init__(self, *args, **kwargs):
        self.app_id = None
        self.auth = None
        self.birthday = None
        self.city = None
        self.country = None
        self.date_registration = None
        self.discount = None
        self.email = None
        self.feed_news = None
        self.feed_sms = None
        self.flowers = None
        self.id = None
        self.ip = None
        self.name = None
        self.phone = None
        self.phone_home = None
        self.phone_work = None
        self.timezone = None

        super(User, self).__init__(*args, **kwargs)


class Page(Object):

    """
    :type id: int
    :type slug: str
    """

    def __init__(self, *args, **kwargs):
        self.application_id = None
        self.category = None
        self.content = None
        self.date = None
        self.file_name = None
        self.id = None
        self.is_file = None
        self.order = None
        self.promo = None
        self.published = None
        self.slug = None
        self.title = None
        self.url = None

        super(Page, self).__init__(*args, **kwargs)


class Review(Object):

    """
    :type id: int
    :type name: str
    """

    def __init__(self, *args, **kwargs):
        self.allow_partners = None
        self.answer = None
        self.app_id = None
        self.avatar_url = None
        self.city = None
        self.city_id = None
        self.comment = None
        self.comment_courier = None
        self.comment_florist = None
        self.comment_florist_good = None
        self.comment_flowers_quality = None
        self.comment_flowers_rating = None
        self.comment_manager = None
        self.comment_operator = None
        self.created = None
        self.email = None
        self.external_url = None
        self.id = None
        self.name = None
        self.order_city = None
        self.order_date = None
        self.order_id = None
        self.phone = None
        self.status = None
        self.urls = None

        super(Review, self).__init__(*args, **kwargs)



from floraconcierge.client.types import Object


class Child(Object):

    """
    """

    def __init__(self, *args, **kwargs):
        self.count = None
        self.product = None

        super(Child, self).__init__(*args, **kwargs)


class Color(Object):

    """
    :type id: int
    """

    def __init__(self, *args, **kwargs):
        self.id = None
        self.title = None

        super(Color, self).__init__(*args, **kwargs)


class Category(Object):

    """
    :type id: int
    :type name: str
    :type parent_id: int
    :type slug: str
    """

    def __init__(self, *args, **kwargs):
        self.id = None
        self.is_hand_order = None
        self.name = None
        self.parent_id = None
        self.slug = None
        self.type = None
        self.weight = None

        super(Category, self).__init__(*args, **kwargs)



from gzip import GzipFile
from StringIO import StringIO

import urllib2


class UserAgentProcessor(urllib2.BaseHandler):
    """A handler to add a custom UA string to urllib2 requests
    """

    def __init__(self, uastring):
        self.handler_order = 100
        self.ua = uastring

    def http_request(self, request):
        request.add_header("User-Agent", self.ua)
        return request

    https_request = http_request


class GZipProcessor(urllib2.BaseHandler):
    """
    A handler to add gzip capabilities to urllib2 requests
    http://techknack.net/python-urllib2-handlers/
    """

    def http_request(self, req):
        req.add_header("Accept-Encoding", "gzip")
        return req

    https_request = http_request

    def http_response(self, req, resp):
        if resp.headers.get("content-encoding") == "gzip":
            gz = GzipFile(
                fileobj=StringIO(resp.read()),
                mode="r"
            )
            old_resp = resp
            resp = urllib2.addinfourl(gz, old_resp.headers, old_resp.url, old_resp.code)
            resp.msg = old_resp.msg

        return resp

    https_response = http_response



from floraconcierge.client.types import Object


class Collection(Object, list):
    def __init__(self, *args, **kwargs):
        super(Collection, self).__init__(*args, **kwargs)

        self.extend(self.items)

    def find(self, **kwargs):
        _matches = len(kwargs)

        for k, v in kwargs.iteritems():
            kwargs[k] = unicode(v).lower()

        for item in self:
            found = 0
            for k, v in kwargs.iteritems():
                if unicode(getattr(item, k)).lower() == v:
                    found += 1

            if found == _matches:
                return item

        return None

    def findall(self, **kwargs):
        _matches = len(kwargs)

        for k, v in kwargs.iteritems():
            kwargs[k] = unicode(v).lower()

        items = []
        for item in self:
            found = 0
            for k, v in kwargs.iteritems():
                if unicode(getattr(item, k)).lower() == v:
                    found += 1

            if found == _matches:
                items.append(item)

        from floraconcierge.mapping import Collection as CollectionObj

        return CollectionObj({
            'count': len(items),
            'items': items,
            'offset': 0,
            'total': len(items)
        })



from floraconcierge.client.types import Object
from floraconcierge.mapping.mixins.model.shop import Currency as CurrencyMixin
from floraconcierge.mapping.mixins.model.shop import Product as ProductMixin


class Order(Object):

    """
    :type id: int
    :type invoices: list of Invoice|Collection
    :type is_sms: str
    :type products: list of Child|Collection
    :type user: User
    """

    def __init__(self, *args, **kwargs):
        self.app_id = None
        self.currency_id = None
        self.date_delive = None
        self.date_order = None
        self.discount = None
        self.id = None
        self.invoices = None
        self.is_sms = None
        self.language = None
        self.order_key = None
        self.other = None
        self.photo = None
        self.products = None
        self.products_free = None
        self.promocode = None
        self.reason = None
        self.receiver_address = None
        self.receiver_city = None
        self.receiver_city_id = None
        self.receiver_country = None
        self.receiver_country_id = None
        self.receiver_email = None
        self.receiver_name = None
        self.receiver_phone = None
        self.sender_city = None
        self.sender_country = None
        self.sender_email = None
        self.sender_gmt = None
        self.sender_ip = None
        self.sender_name = None
        self.sender_phone = None
        self.status = None
        self.status_changed = None
        self.status_plain = None
        self.summ = None
        self.summ_manual = None
        self.time = None
        self.total = None
        self.totals = None
        self.totalusd = None
        self.type = None
        self.user = None
        self.user_id = None
        self.wishes = None

        super(Order, self).__init__(*args, **kwargs)


class Currency(CurrencyMixin):

    """
    :type id: int
    :type name: str
    """

    def __init__(self, *args, **kwargs):
        self.description = None
        self.format = None
        self.id = None
        self.name = None
        self.rounding = None
        self.rounding_direction = None
        self.title = None
        self.usdvalue = None

        super(Currency, self).__init__(*args, **kwargs)


class Paysystem(Object):

    """
    :type id: int
    :type name: str
    :type slug: str
    """

    def __init__(self, *args, **kwargs):
        self.id = None
        self.imageurl = None
        self.name = None
        self.order = None
        self.slug = None
        self.text = None

        super(Paysystem, self).__init__(*args, **kwargs)


class Salon(Object):

    """
    :type id: int
    :type name: str
    :type type: str
    """

    def __init__(self, *args, **kwargs):
        self.address = None
        self.address_link = None
        self.city_id = None
        self.contacts = None
        self.cost = None
        self.cost_usd = None
        self.florist_id = None
        self.id = None
        self.name = None
        self.type = None

        super(Salon, self).__init__(*args, **kwargs)


class Product(ProductMixin):

    """
    :type categories: list|Collection
    :type children: list of Child|Collection
    :type complex: bool
    :type id: int
    :type price_text: str
    :type properties: str
    :type slug: str
    :type weight: int
    """

    def __init__(self, *args, **kwargs):
        self.available = None
        self.categories = None
        self.children = None
        self.comment = None
        self.complex = None
        self.composition = None
        self.compositions_info = None
        self.description = None
        self.gallery = None
        self.id = None
        self.order = None
        self.price = None
        self.price_text = None
        self.prices = None
        self.priceusd = None
        self.properties = None
        self.realprice = None
        self.realprices = None
        self.realpriceusd = None
        self.search_tags = None
        self.slug = None
        self.title = None
        self.type = None
        self.url = None
        self.weight = None

        super(Product, self).__init__(*args, **kwargs)



from floraconcierge.client.types import Object


class Text(Object):

    """
    :type id: int
    :type parent_id: int
    """

    def __init__(self, *args, **kwargs):
        self.id = None
        self.parent_id = None
        self.text = None
        self.title = None

        super(Text, self).__init__(*args, **kwargs)



import calendar
from datetime import datetime, timedelta
import hashlib
import json
import urllib
import urllib2

import floraconcierge
from floraconcierge import errors
from floraconcierge.client import handlers
from floraconcierge.client.types import from_dict
from floraconcierge.errors import get_result_error


class Environment(object):
    def __init__(self):
        self.language = 'en'
        self.country = None
        self.city = None
        self.date = ApiClient.get_timestamp()
        self.currency = 'usd'
        self.ip = '127.0.0.1'
        self.promo = None
        self.user_auth_key = None

    @property
    def params(self):
        return {
            '_lang': self.language or '',
            '_country': self.country or '',
            '_city': self.city or '',
            '_date': self.date or '',
            '_currency': self.currency or '',
            '_ip': self.ip,
            '_promo': self.promo or '',
            '_user_auth_key': self.user_auth_key or ''
        }

    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, state):
        now = ApiClient.get_timestamp()

        if now > state['date']:
            state['date'] = now

        self.__dict__.update(state)


class ApiResult(object):
    def __init__(self, func, params=None, post=None):
        self.func = func
        self.params = params
        self.post = post
        self.url = None
        self.body = None
        self.http_code = None
        self.time = None
        self.size = None

        self._data = None

    def process(self, response):
        try:
            self._data = json.loads(response, object_hook=from_dict)
            self.size = len(response)
        except ValueError, e:
            self.body = response

            raise errors.MalformedJSONResultError(self, e.message)

        if self._data.error.code:
            raise get_result_error(self, self._data.error)

    @property
    def value(self):
        return self._data and self._data.result or None

    def dump(self):
        return self.value


class ApiClient(object):
    def __init__(self, api_id, secret_key):
        self._api_id = api_id
        self._secret_key = secret_key

        self.env = Environment()

        self._opener = urllib2.build_opener(
            handlers.UserAgentProcessor(floraconcierge.HTTP_CLIENT_UA),
            handlers.GZipProcessor()
        )

        self._api_endpoint = 'http://api.floraconcierge.com/v1/'
        self._debug = False
        self._calls = list()
        self._calls_time = timedelta()

        # importing here avoid import recursion
        from floraconcierge.services import Manager

        self.services = Manager(self)

    @property
    def api_endpoint(self):
        return self._api_endpoint

    @api_endpoint.setter
    def api_endpoint(self, url):
        self._api_endpoint = url

    @property
    def debug(self):
        return self._debug

    @debug.setter
    def debug(self, value):
        self._debug = bool(value)

    @staticmethod
    def get_timestamp(time=None):
        time = time or datetime.now()

        return calendar.timegm(time.timetuple())

    @property
    def calls(self):
        return self._calls

    @calls.deleter
    def calls(self):
        self._calls = list()
        self._calls_time = timedelta()

    def sign(self, **kwargs):
        sign = dict((k, v) for k, v in kwargs.iteritems() if k[0] != '_')

        keys = sign.keys()
        keys.sort()

        tosign = []
        for k in keys:
            tosign.append("%s=%s" % (k, sign[k]))

        tosign = "".join(tosign)

        return hashlib.sha1("%s:%s" % (self._secret_key, tosign)).hexdigest()

    def call(self, func, get=None, post=None):
        get, post, start = get or dict(), post or dict(), datetime.now()

        assert isinstance(func, (str, unicode)), 'Supported only string function name'
        assert isinstance(get, dict), 'Supported only dict GET data'
        assert isinstance(post, dict), 'Supported only dict POST data'

        _get, _post = get.copy(), post.copy()

        get = dict((k, v) for k, v in get.iteritems() if v is not None)
        get.update({
            '_sign': self.sign(**get),
            '_api_id': self._api_id,
        })
        get.update(self.env.params)

        url = "%s%s?%s" % (self._api_endpoint, func.lstrip(), urllib.urlencode(get))
        request = urllib2.Request(url)
        if _post:
            request.data = json.dumps(_post)

        try:
            response = self._opener.open(request)
            body = response.read()
        except urllib2.URLError, e:
            raise errors.HTTPError(e, e.reason)

        result = ApiResult(func, _get, _post)
        result.url = response.geturl()
        result.http_code = response.getcode()
        result.time = datetime.now() - start

        if self.debug:
            result.body = body

        self._calls.append(result)
        self._calls_time += result.time

        result.process(body)

        return result.value

    def callstats(self):
        return (self._calls_time, self._calls)


class Service(object):
    """
    @type client: ApiClient
    """

    def __init__(self, client):
        self._client = client

    def _callapi(self, func, get=None, post=None):
        return self._client.call(func, get, post)



from threading import local

from floraconcierge.client import ApiClient

_active = local()


def get_apiclient():
    """
    :rtype: ApiClient
    """
    if not hasattr(_active, "value"):
        raise ValueError('Not activated any apiclient. Please ``activate`` apiclient first')

    client = _active.value
    assert isinstance(_active.value, ApiClient)

    return client


def activate(client):
    """
    Sets the api client for the current thread.

    The ``timezone`` argument must be an instance of a ApiClient subclass
    """
    assert isinstance(client, ApiClient), 'ApiClient object required'
    _active.value = client


def deactivate():
    """
    Unsets the api client for the current thread.
    """
    if hasattr(_active, "value"):
        del _active.value


class override(object):
    """
    Temporarily set the api client for the current thread.
    """

    def __init__(self, client):
        self.client = client
        self.old_client = getattr(_active, 'value', None)

    def __enter__(self):
        if self.client is None:
            deactivate()
        else:
            activate(self.client)

    def __exit__(self, exc_type, exc_value, traceback):
        if self.old_client is None:
            deactivate()
        else:
            _active.value = self.old_client



from floraconcierge.shortcuts import get_apiclient


def get_invoice_by_key(id):
    return get_apiclient().services.invoices.get_invoice_by_key(id)


def get_order_by_key(id):
    return get_apiclient().services.invoices.get_order_by_key(id)


def process_invoice_by_id(id, paysystem_id, amount, postObject):
    return get_apiclient().services.invoices.process_invoice_by_id(
        id,
        paysystem_id,
        amount,
        postObject)



from floraconcierge.shortcuts import get_apiclient


def address_delete(id):
    return get_apiclient().services.users.address_delete(id)


def address_list():
    return get_apiclient().services.users.address_list()


def address_process(postObject):
    return get_apiclient().services.users.address_process(postObject)


def check_password(email, password):
    return get_apiclient().services.users.check_password(email, password)


def exists(email):
    return get_apiclient().services.users.exists(email)


def get_logged_user():
    return get_apiclient().services.users.get_logged_user()


def login(email, password=None, remember=None, checkip=None):
    return get_apiclient().services.users.login(
        email,
        password=password,
        remember=remember,
        checkip=checkip)


def logout():
    return get_apiclient().services.users.logout()


def registration(email):
    return get_apiclient().services.users.registration(email)


def reset_password(email):
    return get_apiclient().services.users.reset_password(email)


def social_list():
    return get_apiclient().services.users.social_list()


def social_login(token):
    return get_apiclient().services.users.social_login(token)


def social_related():
    return get_apiclient().services.users.social_related()


def social_signin(service, redirect_success, redirect_fail):
    return get_apiclient().services.users.social_signin(
        service,
        redirect_success,
        redirect_fail)


def update_info(
        password=None,
        name=None,
        phone=None,
        country=None,
        city=None,
        birthday=None,
        flowers=None):
    return get_apiclient().services.users.update_info(
        password=password,
        name=name,
        phone=phone,
        country=country,
        city=city,
        birthday=birthday,
        flowers=flowers)



from ordered_set import OrderedSet


class LimitedOrderedSet(OrderedSet):
    def __init__(self, limit=0, seq=()):
        setattr(self, 'limit', limit)

        super(LimitedOrderedSet, self).__init__(seq)

        self._check_length()

    def copy(self):
        return LimitedOrderedSet(limit=self.limit, seq=self)

    def discard(self, key):
        self.items.remove(key)
        self.map.pop(key)

    def _check_length(self):
        if not self.limit:
            return

        if len(self) > self.limit:
            for v in self[0:-self.limit]:
                self.remove(v)

    def __getitem__(self, item):
        res = super(LimitedOrderedSet, self).__getitem__(item)

        if isinstance(res, OrderedSet):
            return LimitedOrderedSet(self.limit, list(res))
        else:
            return res

    def __getstate__(self):
        return (self.limit, list(self))

    def __setstate__(self, state):
        self.__init__(*state)

    def add(self, key):
        result = super(LimitedOrderedSet, self).add(key)

        self._check_length()

        return result

    append = add





from debug_toolbar.panels import Panel
from floraconcierge.client import ApiClient, ApiResult


class FloraConciergeRequests(Panel):
    name = 'FloraConciergeRequestsPanel'
    template = 'floraconcierge/debug_toolbar_floraconcierge_requests.html'
    title = 'FloraConcierge Api requests'
    nav_title = 'FloraConcierge'
    has_content = True

    def __init__(self, *args, **kwargs):
        super(FloraConciergeRequests, self).__init__(*args, **kwargs)

        self._num_calls = 0
        self._calls_time = 0

    @property
    def nav_subtitle(self):
        return "%s calls in %.2fms" % (self._num_calls, self._calls_time)

    def process_response(self, request, response):
        api = request.api
        assert isinstance(api, ApiClient)

        call_time, calls = api.callstats()
        self._num_calls = len(calls)
        self._calls_time = call_time.total_seconds()

        queries = []
        width_ratio_tally = 0.0
        for c in calls:
            assert isinstance(c, ApiResult)

            q = {
                'name': c.func,
                'duration': c.time.total_seconds(),
                'width_ratio': c.time.total_seconds() / self._calls_time * 100,
                'params': c.params,
                'trace_color': '#800000',
                'url': c.url,
                'size': c.size,
                'data': c.dump()
            }

            try:
                q['width_ratio'] = (q['duration'] / self._calls_time) * 100
                q['width_ratio_relative'] = (100.0 * q['width_ratio'] / (100.0 - width_ratio_tally))
            except ZeroDivisionError:
                q['width_ratio'] = 0
                q['width_ratio_relative'] = 0

            q['start_offset'] = width_ratio_tally

            width_ratio_tally += q['width_ratio']

            queries.append(q)

        self.record_stats({
            'calls': queries,
            'call_time': call_time.total_seconds()
        })



from floraconcierge.shortcuts import get_apiclient


def set_result(token_id, is_success, data):
    return get_apiclient().services.auth.set_result(token_id, is_success, data)


def get_token(id):
    return get_apiclient().services.auth.get_token(id)



from floraconcierge.shortcuts import get_apiclient


def get_categories_list(parent_id=None):
    return get_apiclient().services.shop.get_categories_list(
        parent_id=parent_id)


def get_category_by_id(id):
    return get_apiclient().services.shop.get_category_by_id(id)


def get_currencies_list(order_id=None):
    return get_apiclient().services.shop.get_currencies_list(order_id=order_id)


def get_currency_by_id(id, order_id=None):
    return get_apiclient().services.shop.get_currency_by_id(
        id,
        order_id=order_id)


def get_gallery():
    return get_apiclient().services.shop.get_gallery()


def calculate_order(postObject):
    return get_apiclient().services.shop.calculate_order(postObject)


def get_order_by_id(id):
    return get_apiclient().services.shop.get_order_by_id(id)


def get_last_delivered():
    return get_apiclient().services.shop.get_last_delivered()


def get_user_orders(is_draft=None):
    return get_apiclient().services.shop.get_user_orders(is_draft=is_draft)


def get_order_reasons_text_by_id(id):
    return get_apiclient().services.shop.get_order_reasons_text_by_id(id)


def get_order_reasons_texts(parent_id=None):
    return get_apiclient().services.shop.get_order_reasons_texts(
        parent_id=parent_id)


def get_order_reason_by_id(id):
    return get_apiclient().services.shop.get_order_reason_by_id(id)


def get_order_reasons():
    return get_apiclient().services.shop.get_order_reasons()


def get_order_status(id):
    return get_apiclient().services.shop.get_order_status(id)


def submit_order_phone(postObject):
    return get_apiclient().services.shop.submit_order_phone(postObject)


def submit_order_request(postObject):
    return get_apiclient().services.shop.submit_order_request(postObject)


def submit_order(postObject):
    return get_apiclient().services.shop.submit_order(postObject)


def submit_order_ticket(id, key, message):
    return get_apiclient().services.shop.submit_order_ticket(id, key, message)


def get_order_tickets(id, key):
    return get_apiclient().services.shop.get_order_tickets(id, key)


def get_payment_system_by_id(id):
    return get_apiclient().services.shop.get_payment_system_by_id(id)


def get_payment_systems_list():
    return get_apiclient().services.shop.get_payment_systems_list()


def get_product_by_id(id):
    return get_apiclient().services.shop.get_product_by_id(id)


def get_featured_products(city_id=None, country_id=None, product_of_day=None):
    return get_apiclient().services.shop.get_featured_products(
        city_id=city_id,
        country_id=country_id,
        product_of_day=product_of_day)


def get_parent_product_by_id(id):
    return get_apiclient().services.shop.get_parent_product_by_id(id)


def search_products(query):
    return get_apiclient().services.shop.search_products(query)


def get_products_list(
        category=None,
        products=None,
        ingredient=None,
        random=None):
    return get_apiclient().services.shop.get_products_list(
        category=category,
        products=products,
        ingredient=ingredient,
        random=random)


def check_promo_code(code):
    return get_apiclient().services.shop.check_promo_code(code)


def get_review_by_id(id):
    return get_apiclient().services.shop.get_review_by_id(id)


def submit_review(postObject):
    return get_apiclient().services.shop.submit_review(postObject)


def get_reviews(limit=None, offset=None, product_id=None):
    return get_apiclient().services.shop.get_reviews(
        limit=limit,
        offset=offset,
        product_id=product_id)


def get_cash_salons_cities(country_id):
    return get_apiclient().services.shop.get_cash_salons_cities(country_id)


def get_cash_salons_countries():
    return get_apiclient().services.shop.get_cash_salons_countries()


def get_cash_salons_by_id(salon_id):
    return get_apiclient().services.shop.get_cash_salons_by_id(salon_id)


def get_cash_salons_by_city(city_id):
    return get_apiclient().services.shop.get_cash_salons_by_city(city_id)



from floraconcierge.client.types import Object


class Signin(Object):

    """
    :type id: int
    """

    def __init__(self, *args, **kwargs):
        self.app_id = None
        self.data = None
        self.expire = None
        self.id = None
        self.is_success = None
        self.redirect_fail = None
        self.redirect_success = None
        self.service = None
        self.token = None
        self.user_id = None

        super(Signin, self).__init__(*args, **kwargs)
